<?php

namespace apexl\Io\modules\authentication\controllers;

use apexl\Io\includes\System;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;

class authenticationController
{
    /**
     * @method GET
     * @return Response
     */
    public function authenticated(Request $request, Response $response)
    {
        return System::asJson($response, ['authenticated' => $request->getAttribute('authenticated', true)]);
    }
}
